/*
:ts=4

    (C) 2000 AROS - The Amiga Research OS
    (C) 2002-2006 Harry Sintonen
    $Id: info.c,v 1.7 2010/12/29 18:10:34 tokai Exp $
 
    Desc:
    Lang: English

*/

/******************************************************************************
 
 
    NAME
 
        Info
 
    SYNOPSIS
 
        DISKS/S, VOLS=VOLUMES/S, GOODONLY/S, BLOCKS/S, VERBOSE/S, DEVICES/M
 
    LOCATION
 
        Workbench:C
 
    FUNCTION
 
    Show information on file system devices and volumes. When given no
    arguments, information on all devices and volumes found in the system
    is displayed. If information is wanted only for some specific devices,
    these names may be given as arguments.
 
    INPUTS
 
    DISKS     --  show information on file system devices
    VOLS      --  show information on volumes
    GOODONLY  --  don't show any information on bad devices or volumes
    BLOCKS    --  show additional block size and usage information
    VERBOSE   --  show additional hardware driver informations for a dos device
    DEVICES   --  device names to show information about
 
    RESULT
 
    NOTES
 
    EXAMPLE
 
  Info
 
  Unit                 Size    Used    Free Full Errs   State    Type Name
  Foreign harddisk:  964.1M  776.7M  187.4M  81%    0 read/write  OFS Workbench
  RAM:                 8.0M    7.1M    7.1M  12%    0 read/write  OFS Ram Disk
 
    BUGS
 
    SEE ALSO
 
    INTERNALS
 
    The original source showed that AROS version of ReadArgs() handles
    the /M switch with zero arguments differently from AmigaOS. While AROS
    returns an array where the first pointer is NULL, AmigaOS just returns
    NULL.
 
    HISTORY
 
    16.11.2000  SDuvan  --  converted to AROS
    23.12.2000  SDuvan  --  changed semantics and updated
                            (now fully functional)
    19.12.2002  Piru    --  tweaked and improved. better error reporting.
                            removed three deadly bugs (jmp to 0 and 2 x
                            memtrash). Removed DTF_SUBST from volume date
                            display. Use unsigned formatting for BLOCKS/S.
                            Bumped to 50.1.
    29.12.2002  Piru    --  fixed some nasty hit with BSTR ffsm.
                            Bumped to 50.2.
 
    Based on the original by:
     1997-1998 by Stephan Rupprecht
    All rights resevered
*/

#define  DEBUG  0

#ifdef __MORPHOS__
#include "info_morphos1.h"
#endif
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/exall.h>
#include <dos/filehandler.h>
#include <exec/memory.h>
#include <libraries/locale.h>

#include <aros/debug.h>

#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/utility.h>
#include <proto/locale.h>
#include <proto/alib.h>

#include <string.h>

#undef D
#undef bug
#define	D(x)	;
#define bug	dprintf

/*
 * could be optimized for rawdofmt somehow
 */
static const ULONG	RawFunc[1]=
    {
        0x16c04e75
    };
#define __sprintf(buf, string, tags...) \
({ULONG _tags[] = { tags }; RawDoFmt(string, (APTR)_tags, (void*) &RawFunc, buf);})

#define ID_MAC_DISK2       (0x4d414300L) /* MAC\0 - xfs mac disk */
#define ID_MNX1_DISK       (0x4d4e5801L) /* MNX\1 - xfs minix disk */
#define ID_MNX2_DISK       (0x4d4e5802L) /* MNX\2 - xfs minix disk */
#define ID_QL5A_DISK       (0x514c3541L) /* QL5A - xfs ql 720k / ed disk */
#define ID_QL5B_DISK       (0x514c3542L) /* QL5B - xfs ql 1440k disk */
#define ID_ZXS0_DISK       (0x5a585300L) /* Spectrum Disciple - xfs */
#define ID_ZXS1_DISK       (0x5a585301L) /* Spectrum UniDos - xfs */
#define ID_ZXS2_DISK       (0x5a585302L) /* Spectrum SamDos - xfs */
#define ID_ZXS4_DISK       (0x5a585304L) /* Spectrum Opus 180k - xfs */
#define ID_ARME_DISK       (0x41524d44L) /* Archimedes - xfs */
#define ID_ARMD_DISK       (0x41524d43L) /* Archimedes - xfs */
#define ID_CPM_DISK        (0x43505c4dL) /* CP/M - xfs */
#define ID_ZXS3_DISK       (0x5a585303L) /* ZXS\3 - Plus3Dos xfs */
#define ID_1541_DISK       (0x31353431L) /* 1541 - xfs */
#define ID_1581_DISK       (0x31353831L) /* 1581 - xfs */
#define ID_MAC_DISK        (0x4d534800L) /* MSH\0 - CrossDos MACDisk ?! */
#define ID_ACD0_DISK       (0x41434400L) /* ACD\0 - AmiCDFS disk */
#define ID_CACHECDFS_DISK  (0x43443031L)
#define ID_ASIMCDFS_DISK   (0x662dabacL)
#define ID_PFS2_DISK       (0x50465302L)
#define ID_PFS2_SCSI_DISK  (0x50445300L)
#define ID_PFS2_muFS_DISK  (0x6d755046L)
#define ID_FLOPPY_PFS_DISK (0x50465300L)
#define ID_P2A0_DISK       (0x50324130L)
#define ID_AFS0_DISK       (0x41465300L) /* AFS\0 */
#define ID_muFS_DISK       (0x6d754653L) /* muFS - Mulituserfsys */


/* Prototypes */

static
void FmtProcedure(void);
static inline ULONG ExtUDivMod32(ULONG a, ULONG b, ULONG *mod)
{
	*mod = a % b;

	return a / b;
}
static
int doInfo();

const struct EmulLibEntry FmtProcedure_gate =
    {
	    TRAP_LIBNR, 0, (void (*)(void))FmtProcedure
    };

#include "info_version.h"

static const char version[] = VERSTAG;

struct Catalog	*cat;
struct Locale   *loc = NULL;
ULONG            MaxLen;

APTR             Pool;


/* catalog string id:s */
enum
{
	UNIT,
	DEVTITLE,
	DISKSTITLE,
	DEVFMTSTR,
	DATEFMTSTR,
	READONLY,
	READWRITE,
	VALIDATING,
	MOUNTEDSTR,
	SMALLNUMFMT,
	BIGNUMFMT,
	VOLNAMEFMTSTR,
	BLOCKSSTR,
	DOSTYPEFMTSTR
};


#define MAXNAME 108

struct InfoDosNode
{
	struct InfoDosNode *Next;
	ULONG		IsVolume;
	ULONG		DosType;
	struct MsgPort     *Task;
	struct DateStamp    VolumeDate;
	void		*Startup;
	TEXT		Name[MAXNAME];
};

struct InfoDosNode *head = NULL;


struct DiskTypeList
{
	ULONG   id;
	STRPTR  str;
};

struct DiskTypeList dtl[] =
    {
	    {
		    ID_DOS_DISK,         "OFS"
	    },
	    { ID_FFS_DISK,         "FFS" },
	    { ID_INTER_DOS_DISK,   "OFS-INT" },
	    { ID_INTER_FFS_DISK,   "FFS-INT" },
	    { ID_FASTDIR_DOS_DISK, "OFS-DC" },
	    { ID_FASTDIR_FFS_DISK, "FFS-DC" },
	    { ID_LNFS_DOS_DISK,    "OFS-LNFS" },
	    { ID_LNFS_FFS_DISK,    "FFS-LNFS" },
	    { ID_MSDOS_DISK,       "MS-DOS" },
	    { ID_ACD0_DISK,        "CDFS" },
	    { ID_CACHECDFS_DISK,   "CDFS" },
	    { ID_ASIMCDFS_DISK,    "CDFS" },
	    { ID_NOT_REALLY_DOS,   "NO DOS" },
	    { ID_MAC_DISK2,        "Mac" },
	    { ID_MNX1_DISK,        "Minix" },
	    { ID_QL5A_DISK,        "QL720k" },
	    { ID_QL5B_DISK,        "QL1.4M" },
	    { ID_CPM_DISK,         "CP/M" },
	    { ID_ZXS3_DISK,        "+3Dos" },
	    { ID_ZXS0_DISK,        "Disciple " },
	    { ID_ZXS1_DISK,        "UniDos" },
	    { ID_ZXS2_DISK,        "SamDos" },
	    { ID_ZXS4_DISK,        "Opus" },
	    { ID_P2A0_DISK,        "NETWORK" },
	    { 0x53465300,          "SFS" },
	    { 0x41465300,          "AFS" },
	    { 0x50465300,          "PFS" },
	    { 0x42464653,          "BFFS" },
	    { ID_CDFS_DISK,        "CDFS" },
	    { ID_CDROM_ISO_DISK,   "CD-ISO" },
	    { ID_CDROM_HSF_DISK,   "CD-HSF" },
	    { ID_CDROM_CDDA_DISK,  "CDDA" },
	    { 0x45585402,          "Ext2" },
	    { 0x58543203,          "Ext3" },
	    { 0x4E544653,          "NTFS" },
	    { 0x58465300,          "SGI-XFS" },
	    { 0x4846532B,          "Mac-HFS+" },
	    { 0x4D414300,          "Mac-HFS" },
	    { 0L, NULL }
    };


/****************************************************************************/

struct ExecBase	 	*SysBase;
struct DosLibrary	*DOSBase;
struct Library   	*UtilityBase;
struct Library   	*LocaleBase;

int main(void)
{
	int res = RETURN_FAIL;

	SysBase=*((struct ExecBase**) 4);

	if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library", 37)))
	{
		if ((UtilityBase=OpenLibrary("utility.library", 37)))
		{
			if ((LocaleBase=OpenLibrary("locale.library", 38)))
			{
				#if 0
				/*  XXX: Makes no sense to waste a catalog with a strange name
				 *       per command. There needs to be a single "sys/commands.catalog"
				 *       with a string range for each command. Disabled for now.
				 *       - tokai
				 */
				static const struct TagItem loctags[] =
				{
					{OC_Version, 2},
					{ TAG_END   , 0 }
				};
				cat = OpenCatalogA(NULL, "info_com.catalog", (struct TagItem *) loctags);
				#endif
				loc = OpenLocale(NULL);
			}
			D(bug("Calling doInfo()\n"));

			res = doInfo();

			if (res != RETURN_OK)
			{
				PrintFault(IoErr(), NULL);
			}

			if (LocaleBase)
			{
				CloseLocale(loc);
				CloseCatalog(cat);
				CloseLibrary(LocaleBase);
			}

			CloseLibrary(UtilityBase);
		}

		CloseLibrary((struct Library *) DOSBase);
	}

	return res;
}

static
STRPTR GetStrFromCat(ULONG id, STRPTR def)
{
	if (cat)
	{
		def = GetCatalogStr(cat, id, def);
	}

	return def;
}

#if 1
#define LPrintf(id, def, tags...) \
({ULONG _tags[] = { tags }; VPrintf(GetStrFromCat(id, def), (APTR)_tags);})
#else
static
void LPrintf(ULONG id, STRPTR def, ...)
{
	def = GetStrFromCat(id, def);

	VPrintf(def, ((IPTR *)(&def))+1);
}
#endif


static
BOOL myMatchPatternNoCase(STRPTR *array, CONST_STRPTR str)
{
	if (array && *array)
	{
		while (*array)
		{
			UBYTE  *p = *array++;
			UBYTE   len = strlen(p);
			UBYTE   matchstr[(len + 2) * 2 + 2];
			UBYTE   name[len + 2];

			if (p[len - 1] != ':')
			{
				CopyMem(p, name, len);
				name[len] = ':';
				name[len + 1] = 0;
				p = name;
			}

			if (ParsePatternNoCase(p, matchstr, sizeof(matchstr)) != -1)
			{
				if (MatchPatternNoCase(matchstr, (STRPTR) str))
				{
					return TRUE;
				}
			}
		}

		return FALSE;
	}

	return TRUE;
}


static
LONG insertnode(struct DosList *dl, struct DosList *ndl, CONST_STRPTR name, STRPTR *strray);

static
BOOL ScanDosList(STRPTR *filter)
{
	struct DosList     *ndl, *dl;
	STRPTR             *strray = NULL;
	BOOL                err = FALSE;
	const STRPTR        emptyfilter[1] = {NULL};

	D(bug("Entered ScanDosList()\n"));

	if (!filter)
	{
		filter = (STRPTR *)emptyfilter;
	}

	if (*filter)
	{
		strray = AllocPooled(Pool, sizeof(STRPTR) * MAX_MULTIARGS);
		if (!strray)
		{
			SetIoErr(ERROR_NO_FREE_STORE);

			return FALSE;
		}

		{

			STRPTR  *p = filter;
			LONG     i = 0;

			while (*p && i < (MAX_MULTIARGS - 1))
				strray[i++] = *p++;

			while (i < MAX_MULTIARGS)
				strray[i++] = NULL;
		}
	}

	D(bug("ScanDosList: unlock\n"));

	/* lock list of devices & vols */
	dl = ndl = LockDosList(LDF_ASSIGNS | LDF_VOLUMES | LDF_DEVICES | LDF_READ);

	D(bug("ScanDosList: scan volume/dirs\n"));

	if (strray)
	{
		STRPTR *p = strray;

		while (*p)
			p++;


		while ((ndl = NextDosEntry(ndl, LDF_ASSIGNS | LDF_VOLUMES)))
		{
			UBYTE   name[((UBYTE *)BADDR(ndl->dol_Name))[0] + 2];
			STRPTR  taskName = "";

			D(bug("ScanDosList: %b\n",ndl->dol_Name));

			__sprintf(name, "%b:", (ULONG) ndl->dol_Name);

			if ((ndl->dol_Type > DLT_VOLUME) || !(myMatchPatternNoCase(strray, name)))
				continue;

			switch(ndl->dol_Type)
			{
			case DLT_VOLUME:
				taskName =  BADDR(ndl->dol_Name)+1; //ndl->dol_DevName;   // ((struct Task *)ndl->dol_Task->mp_SigTask)->tc_Node.ln_Name;

				D(bug("Found volume %s\n", taskName));
				break;

			case DLT_DIRECTORY:
				{
					struct AssignList *al = ndl->dol_misc.dol_assign.dol_List;

					taskName = BADDR(ndl->dol_Name)+1; // ((struct Task *)((struct FileLock *)BADDR(ndl->dol_Lock))->fl_Task->mp_SigTask)->tc_Node.ln_Name;

					D(bug("Found directory %s\n", taskName));

					while (al)
					{
#warning "TODO: Hmm, should fix this?"
						*p++ = ""; // TODO!!!  ((struct Task *)((struct FileLock *)BADDR(al->al_Lock))->fl_Task->mp_SigTask)->tc_Node.ln_Name;
						al = al->al_Next;
					}
				}
				break;
			}

			*p++ = taskName;
		}

		D(bug("ScanDosList: done\n"));

	}
	else
		strray = filter;

	ndl = dl;

	D(bug("ScanDosList: scan volumes/devices\n"));

	while ((ndl = NextDosEntry(ndl, LDF_VOLUMES | LDF_DEVICES)))
	{
		UBYTE  type = ndl->dol_Type;
		UBYTE  name[((UBYTE *)BADDR(ndl->dol_Name))[0] + 2];

		if (type == DLT_DEVICE && (!ndl->dol_Task))
		{
			/*
			 * i had some "broken?" mountlist entry name audio: which
			 * wasn't active but when the filesystem checked failed on
			 * it, info was hanging there.
			 */
			continue;
		}

		__sprintf(name, "%b:", (ULONG) ndl->dol_Name);

		D(bug("Found name %s\n", BADDR(ndl->dol_Name)+1));//ndl->dol_DevName));

		if ((type == DLT_DEVICE) && (myMatchPatternNoCase(strray, name) == FALSE))
		{
			int i;

			D(bug("Failure! -- name = %s, strray = %p\n", name, (void *)strray));

			for (i = 0; strray[i]; i++)
			{
				D(bug("Strray %i = %s\n", i, strray[i]));
			}

			continue;
		}

		if (!insertnode(dl, ndl, name, strray))
		{
			err = TRUE;
			break;
		}
	}

	/* unlock list of devices and volumes */
	UnLockDosList(LDF_ASSIGNS | LDF_VOLUMES | LDF_DEVICES | LDF_READ);

	D(bug("ScanDosList: unlock\n"));

	// strray freed at DeletePool

	return !err;
}


static
LONG insertnode(struct DosList *dl, struct DosList *ndl, CONST_STRPTR name, STRPTR *strray)
{
	struct InfoDosNode *idn;
	UBYTE  type = ndl->dol_Type;
	UBYTE  len = 0;

	/* Avoid dupes - Piru */
	for (idn = head; idn; idn = idn->Next)
	{
		if (strcmp(name, idn->Name) == 0)
		{
			return TRUE;
		}
	}

	idn = (struct InfoDosNode *)AllocPooled(Pool, sizeof(struct InfoDosNode));
	if (!idn)
	{
		SetIoErr(ERROR_NO_FREE_STORE);
		return FALSE;
	}

	//	idn->Task     = (struct MsgPort *)ndl->dol_Task;
	idn->IsVolume = type == DLT_VOLUME;

	while ((idn->Name[len] = name[len]))
	{
		len++;
	}

	if (type == DLT_VOLUME)
	{
	    struct MsgPort *handler;
		handler = ndl->dol_Task;

		if (handler && myMatchPatternNoCase(strray, name))
		{
			/* Scan the devicelist for entry with a matching handler */

			struct DosList *dosentry = dl;

			while ((dosentry = NextDosEntry(dosentry, LDF_DEVICES)))
			{
				if (dosentry->dol_Task == handler)
				{
					UBYTE devname[((UBYTE *)BADDR(dosentry->dol_Name))[0] + 2];

					__sprintf(devname, "%b:", (ULONG) dosentry->dol_Name);

					if (!insertnode(dl, dosentry, devname, strray))
					{
						return FALSE;
					}

					break;
				}
			}
		}

		idn->VolumeDate = ((struct DeviceList *)ndl)->dl_VolumeDate;
		idn->Name[len - 1] = '\0';       /* remove ':' */
		idn->DosType = ((struct DeviceList *)ndl)->dl_DiskType;
		idn->Startup = NULL;
	}
	else if (type == DLT_DEVICE)
	{
		struct FileSysStartupMsg *fssm = (struct FileSysStartupMsg *)BADDR(ndl->dol_misc.dol_handler.dol_Startup);

		idn->DosType = ID_DOS_DISK;

		//  DLT_DEVICE
		if (len > MaxLen)
			MaxLen = len;
#if 1
		if ((ULONG) fssm > 0x400)
		{
			idn->Startup = fssm;
			if ((fssm->fssm_Unit & 0xff000000) == 0 /*||
			    *(UBYTE *)BADDR(fssm->fssm_Device) != 0*/)
			{
				struct DosEnvec *de = (struct DosEnvec *)BADDR(fssm->fssm_Environ);

				if (de && (de->de_TableSize >= DE_DOSTYPE))
				{
					if (de->de_DosType)
						idn->DosType = de->de_DosType;
				}
			}
		}
		else
			idn->Startup = NULL;
#endif

	}

	/* kinda insert sort */
	{
		struct InfoDosNode *work = head;
		struct InfoDosNode *prev = NULL;

		while (work && (Stricmp(idn->Name, work->Name) > 0))
		{
			prev = work;
			work = work->Next;
		}

		if (prev)
			prev->Next = idn;
		else
			head = idn;

		idn->Next = work;
	}

	return TRUE;
}


static
void PrintNum(UQUAD q_num)
{
	/* MBytes ? */
	if (q_num >= 1024ULL)
	{
		ULONG  num, x, xx;
		UBYTE  fmt;

		/* PBytes ? */
		if (q_num >= 0x100000ULL * 1024 * 1024)
		{
			num = q_num >> 30;
			fmt = 'P';
		}
		/* TBytes ? */
		else if (q_num >= 0x100000ULL * 1024)
		{
			num = q_num >> 20;
			fmt = 'T';
		}
		/* GBytes ? */
		else if (q_num >= 0x100000ULL)
		{
			num = q_num >> 10;
			fmt = 'G';
		}
		else /* MBytes */
		{
			num = q_num;
			fmt = 'M';
		}

		num = ExtUDivMod32((num * 100) >> 10, 100, &x);

		/* round */
		x = ExtUDivMod32(x, 10, &xx);

		if (xx >= 5)
		{
			if (++x > 9)
			{
				x = 0;
				num++;
			}
		}

		LPrintf(BIGNUMFMT, "%6ld.%ld%lc", num, x, fmt);
	}
	else
	{
		LPrintf(SMALLNUMFMT, "%8ldK", (ULONG)q_num);
	}
}

static
STRPTR GetFSysStr(ULONG DiskType)
{
	struct DiskTypeList	*dtlptr = dtl;

	STRPTR ptr = NULL;

	do
	{
		if (dtlptr->id == DiskType)
		{
			ptr = dtlptr->str;
			break;
		}
	}
	while (*((ULONG *)dtlptr++));

	if (!ptr)
	{
		static TEXT buffer[5];

		ptr = (STRPTR)buffer;
		*((ULONG *)ptr) = DiskType;

		if (ptr[3] < ' ')
			ptr[3] += '0';

		ptr[4] = '\0';
	}

	return ptr;
}

static
STRPTR GetDevStr(struct FileSysStartupMsg *StartupMsg)
{
	static UBYTE Buffer[256+12];

	if ((StartupMsg->fssm_Unit & 0xff000000)==0)
	{
		UBYTE	*Device;
		//int	len;

		Device	=	BADDR(StartupMsg->fssm_Device);
		if (Device)
		{
			__sprintf(Buffer,"%s : %ld",(ULONG)&Device[1],StartupMsg->fssm_Unit);
			return(Buffer);
		}
	}

	return(NULL);
}

static
STRPTR GetStartupStr(UBYTE *StartupName)
{
	static UBYTE Buffer[256+1];

	memcpy(Buffer,&StartupName[1],StartupName[0]);
	Buffer[StartupName[0]]	=	'\0';
	return(Buffer);
}

enum
{
    ARG_DISKS,
    ARG_VOLS,
    ARG_GOODONLY,
    ARG_BLOCKS,
    ARG_VERBOSE,
    ARG_DEVS,
    NOOFARGS
};


static
int	doInfo()
{
	struct RDArgs       *rdargs=NULL;
	struct Process      *proc;
	struct Window       *win;
	struct InfoDosNode  *idn;
	LONG		res = RETURN_FAIL;
	UBYTE stack_id[sizeof(struct InfoData) + 3];
	struct InfoData *id = (struct InfoData *) (((ULONG) stack_id + 3) & ~3);

	IPTR   args[] = { (IPTR)FALSE,
	                  (IPTR)FALSE,
	                  (IPTR)FALSE,
	                  (IPTR)FALSE,
	                  (IPTR)FALSE,
	                  (IPTR)NULL };

	STRPTR unit = GetStrFromCat(UNIT, "Unit");

	Pool = CreatePool(MEMF_ANY, 1024, 1024);
	if (!Pool)
	{
		SetIoErr(ERROR_NO_FREE_STORE);
		return RETURN_FAIL;
	}

	/* avoid requesters */
	proc = (struct Process *)FindTask(NULL);
	win  = (struct Window *)proc->pr_WindowPtr;
	proc->pr_WindowPtr = (struct Window *) -1;

	D(bug("Calling ReadArgs()\n"));

	/* read arguments */
	rdargs = ReadArgs("DISKS/S,VOLS=VOLUMES/S,GOODONLY/S,BLOCKS/S,VERBOSE/S,DEVICES/M",
	                  args, NULL);

	if (rdargs)
	{
		BOOL     disks    = (BOOL)args[ARG_DISKS];
		BOOL     vols     = (BOOL)args[ARG_VOLS];
		BOOL     goodOnly = (BOOL)args[ARG_GOODONLY];
		BOOL     blocks   = (BOOL)args[ARG_BLOCKS];
		BOOL     verbose  = (BOOL)args[ARG_VERBOSE];
		STRPTR  *devs     = (STRPTR *)args[ARG_DEVS];

		/* If nothing is specified, show everything we got */
		if ((!devs || !*devs) && !disks && !vols)
		{
			vols = TRUE;
			disks = TRUE;
		}

		/* check pattern strings */

		if (devs && *devs)
		{
			STRPTR  *p;

			for (p = devs; (*p); p++)
			{
				TEXT  matchstr[128];

				if (ParsePatternNoCase(*p, matchstr, sizeof(matchstr)) == -1)
				{
					res = RETURN_ERROR;
					goto end;
				}
			}
		}

		MaxLen = strlen(unit);

		D(bug("Calling ScanDosList()\n"));

		/* scan doslist */
		if (ScanDosList(devs))
		{
			STRPTR  dstate[3] = { GetStrFromCat(READONLY,   "read only"),
			                      GetStrFromCat(VALIDATING, "validating"),
			                      GetStrFromCat(READWRITE,  "read/write") };
			STRPTR  datetimeFmt = NULL;
			BOOL    first = TRUE;
			TEXT    nfmtstr[16];
			TEXT    buf[64];

			res = RETURN_WARN;

			D(bug("Printing stuff\n"));

			/* get datetimefmt string */
			if (loc && (GetVar("info_datetime", buf, sizeof(buf), 0L) > 0L))
			{
				datetimeFmt = buf;
			}

			/* calc format string for 'Unit' */
			__sprintf(nfmtstr, "%%-%lds", MaxLen);

			/* show device infomation */
			if ((devs && *devs) || disks || !vols)
			{
				for (idn = head; idn; idn = idn->Next)
				{
					BPTR    lock;
					STRPTR  name = idn->Name;
					BOOL    err;
					LONG    ioerr = 0;

					D(bug("Got name = %s\n", name));

					D(bug("Check FileSystem..IsVolume 0x%lx\n", name,idn->IsVolume));

					if (!idn->IsVolume && IsFileSystem(name))
					{
						D(bug("Check FileSystem\n", name));
						/* if first device to print, print title */
						if (first || blocks)
						{
							if (!first)
								Printf("\n");

							D(bug("Printing device\n"));

							LPrintf(~0, nfmtstr, (ULONG)unit);
							LPrintf(DEVTITLE, "     Size     Used     Free Full Errs   State        Type Name\n");

							first = FALSE;
						}

						err = FALSE;

						D(bug("Locking \"%s\"\n", name));
						lock = Lock(name, SHARED_LOCK);
						D(bug("Lock = %p\n", lock));

						if (lock)
						{
							D(bug("Got lock on %s\n", name));

							memclr(id,sizeof(*id));

							if (Info(lock, id))
							{
								UQUAD  numblocks, numblocksused;
								LONG   gotnumblocks = 0, gotnumblocksused = 0;
								UQUAD  sizekb, usedkb;
								ULONG  x, y;
								UBYTE *DiskStateString;

								res = RETURN_OK;

								D(bug("Got info on %s\n", name));

#ifdef __MORPHOS__
								#include "info_morphos1.c"
#endif
								if (!gotnumblocks)
								{
									numblocks = (ULONG) id->id_NumBlocks;
								}
								if (!gotnumblocksused)
								{
									numblocksused = (ULONG) id->id_NumBlocksUsed;
								}

#if 0
								Printf("  NumSoftErrors %ld UnitNumber %ld DiskState 0x%lx\n",id->id_NumSoftErrors,id->id_UnitNumber,id->id_DiskState);
								Printf("  Numblocks %ld NumBlocksUsed %ld BytesperBlock %ld\n",id->id_NumBlocks, id->id_NumBlocksUsed, id->id_BytesPerBlock);
								Printf("  FreeNumblocks %ld\n",id->id_NumBlocks - id->id_NumBlocksUsed);
								Printf("  DiskType 0x%lx VolumeNode 0x%lx InUse %ld\n",id->id_DiskType,id->id_VolumeNode,id->id_InUse);
#endif

								LPrintf(~0, nfmtstr, (ULONG)name);

								sizekb = (numblocks * id->id_BytesPerBlock) >> 10;
								usedkb = (numblocksused * id->id_BytesPerBlock) >> 10;

								/* Size */
								PrintNum(sizekb);

								/* Used */
								PrintNum(usedkb);

								/* Free */
								PrintNum(sizekb - usedkb);

								D(bug("Calling NameFromLock()\n"));

								if (NameFromLock(lock, name, MAXNAME))
								{
									LONG len = strlen(name) - 1;

									if (name[len] == ':')
									{
										name[len] = '\0';
									}
								}

								if (sizekb >= 0x100000ULL * 1024 * 1024)
								{
									x = sizekb >> 30;
									y = usedkb >> 30;
								}
								else if (sizekb >= 0x100000ULL * 1024)
								{
									x = sizekb >> 20;
									y = usedkb >> 20;
								}
								else if (sizekb >= 0x100000ULL)
								{
									x = sizekb >> 10;
									y = usedkb >> 10;
								}
								else
								{
									x = sizekb;
									y = usedkb;
								}

								if (x)
								{
									x = ExtUDivMod32(y * 1000 / x, 10, &y);

									if (y >= 5)
										x++;
								}
								else
								{
									x = 0;
								}

								// y = ((struct DeviceList *)BADDR(id->id_VolumeNode))->dl_DiskType;

								//				if (!y)
								y = id->id_DiskType;

								if ((idn->DosType & ID_DOS_DISK) != ID_DOS_DISK)
									y = idn->DosType;

								D(bug("DosType = 0x%lx\n", idn->DosType));

								if ((id->id_DiskState >= ID_WRITE_PROTECTED) && (id->id_DiskState <= ID_VALIDATED))
								{
									DiskStateString = dstate[id->id_DiskState - ID_WRITE_PROTECTED];
								}
								else
								{
									DiskStateString = "";
								}
								LPrintf(DEVFMTSTR, "%4ld%% %4ld %-11s%8s %s\n",
								        x,
								        id->id_NumSoftErrors,
								        (ULONG)DiskStateString,
								        (ULONG)GetFSysStr(y),
								        (ULONG)name);

								if (verbose && idn->Startup)
								{
									UBYTE	*DevString;
									UBYTE	*StartupString;
									if ((DevString=GetDevStr((struct FileSysStartupMsg *) idn->Startup)))
									{
										Printf("  -> %s\n",
										       (ULONG)DevString);
									}
									else if ((StartupString=GetStartupStr((UBYTE*) idn->Startup)))
									{
										Printf("  -> %s\n",
										       (ULONG)StartupString);
									}
								}

								if (blocks)
								{
									LPrintf(BLOCKSSTR,
									        "\nTotal blocks: %-10llu  Blocks used: %llu\n"
									        " Blocks free: %-10llu    Blocksize: %lu\n",
									        (ULONG) (numblocks >> 32), (ULONG) numblocks,
									        (ULONG) (numblocksused >> 32), (ULONG) numblocksused,
									        (ULONG) ((numblocks - numblocksused) >> 32), (ULONG) (numblocks - numblocksused),
									        id->id_BytesPerBlock );
								}
							}
							else
							{
								ioerr = IoErr();
								err = TRUE;
								D(bug("Info failure\n"));
							}

							UnLock(lock);
						}
						else
						{
							ioerr = IoErr();
							err = TRUE;
						}

						if (err && ioerr && !goodOnly)
						{
							LPrintf(~0, nfmtstr, (ULONG)name);
							PrintFault(ioerr, "\t");
						}
					}
					if (SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
					{
						SetIoErr(ERROR_BREAK);
						res = RETURN_FAIL;
						goto end;
					}
				}
			}

			//SetIoErr(0);

			/* show volumes */
			if (vols || (!devs && !disks))
			{
				if (!first)
					PutStr("\n");

				LPrintf(DISKSTITLE, "Volumes available:\n");

				for (MaxLen = 15, idn = head; idn; idn = idn->Next)
				{
					if (idn->IsVolume)
					{
						LONG len = strlen(idn->Name);

						if (len > MaxLen)
							MaxLen = len;
					}
				}

				__sprintf(nfmtstr, "%%-%lds%%-10s", MaxLen+1);

				for (idn = head; idn; idn = idn->Next)
				{
					if (idn->IsVolume)
					{
						LPrintf(VOLNAMEFMTSTR, nfmtstr, (ULONG)idn->Name,
						        (ULONG)GetStrFromCat(MOUNTEDSTR, "[Mounted]"));
						// idn->Task ? GetStrFromCat(MOUNTEDSTR, "[Mounted]") : ""); TODO

						if (datetimeFmt)
						{
							UBYTE datestr[128];
							static struct Hook hook;

							hook.h_Entry    = (HOOKFUNC) &FmtProcedure_gate;
							hook.h_Data     = datestr;

							FormatDate(loc, datetimeFmt, &idn->VolumeDate, &hook);

							PutStr(datestr);
						}
						else
						{
							TEXT  StrDay[LEN_DATSTRING];
							TEXT  StrDate[LEN_DATSTRING];
							TEXT  StrTime[LEN_DATSTRING];

							struct DateTime dt;

							dt.dat_Flags   = 0; // DTF_SUBST  - Better without, imho. - Piru
							dt.dat_Format  = FORMAT_DOS;
							dt.dat_StrDay  = StrDay;
							dt.dat_StrDate = StrDate;
							dt.dat_StrTime = StrTime;
							dt.dat_Stamp   = idn->VolumeDate;

							if (DateToStr(&dt))
							{
								if (Strnicmp(StrDate, StrDay, strlen(StrDay)) == 0)
								{
									dt.dat_Flags = 0L;
									DateToStr(&dt);
								}

								LPrintf(DATEFMTSTR, "created %11s, %-10s %s",
								        (ULONG)StrDay, (ULONG)StrDate, (ULONG)StrTime);
							}
						}
						if (idn->DosType)
						{
							LPrintf(DOSTYPEFMTSTR," <%s>",(ULONG)GetFSysStr(idn->DosType));
						}
						PutStr("\n");
					}

					if (SetSignal(0L, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
					{
						SetIoErr(ERROR_BREAK);
						res = RETURN_FAIL;
						goto end;
					}
				}
			}

			//res = RETURN_OK;
		}

end:
		D(bug("end:\n"));

		/* free args */
		FreeArgs(rdargs);
	}

	/* reset window pointer of our process */

	proc->pr_WindowPtr = win;

	/* free allocated memory */

	DeletePool(Pool);

	return res;
}


static
void FmtProcedure(void)
{
	struct Hook *hook     = (struct Hook *) REG_A0;
	UBYTE a               = (UBYTE) REG_A1;
	//struct Locale *locale = (struct Locale *) REG_A2;

	*((STRPTR)hook->h_Data) = a;
	((STRPTR)hook->h_Data)++;
}
